// checkpwd.cpp
//
// Validation of user credentials.
//
// $Id: $
//

#include "stdafx.h"
#include "checkpwd.h"
#include "sspiauth.h"

#include <lm.h>

//---------------------------------------------------------------------------
// CheckPassword_LogonUser
//
//  Validates user credentials using LogonUser function.
//
//  Parameters:
//	  pszDomainName - domain name
//	  pszUserName   - user name
//	  pszPassword	- password
//	  phToken       - pointer to a variable that receives user token
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
CheckPassword_LogonUser(
	IN PCTSTR pszDomainName,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPassword,
	OUT PHANDLE phToken
	)
{
	_ASSERTE(pszUserName != NULL);
	_ASSERTE(pszPassword != NULL);

	HANDLE hToken;

	TCHAR szDomainName[DNLEN + 1];
	TCHAR szUserName[UNLEN + 1];
	TCHAR szPassword[PWLEN + 1];

	if (pszDomainName == NULL)
	{
		BYTE bSid[8 + 4 * SID_MAX_SUB_AUTHORITIES];
		ULONG cbSid = sizeof(bSid);
		ULONG cchDomainName = countof(szDomainName);
		SID_NAME_USE Use;

		if (!LookupAccountName(NULL, pszUserName, (PSID)bSid, &cbSid,
							   szDomainName, &cchDomainName, &Use))
			return FALSE;
	}
	else
		lstrcpyn(szDomainName, pszDomainName, countof(szDomainName));

	lstrcpyn(szUserName, pszUserName, countof(szUserName));
	lstrcpyn(szPassword, pszPassword, countof(szPassword));

	if (!LogonUser(szUserName, szDomainName, szPassword, 
				   LOGON32_LOGON_NETWORK, LOGON32_PROVIDER_DEFAULT, 
				   &hToken))
		return FALSE;

	if (phToken == NULL)
		_VERIFY(CloseHandle(hToken));
	else
		*phToken = hToken;

	return TRUE;
}

//---------------------------------------------------------------------------
// CheckPassword_SSPI
//
//  Validates user credentials using SSPI interface.
//
//  Parameters:
//	  pszDomainName - domain name
//	  pszUserName   - user name
//	  pszPassword	- password
//	  phToken       - pointer to a variable that receives user token
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
CheckPassword_SSPI(
	IN PCTSTR pszDomainName,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPassword,
	OUT PHANDLE phToken
	)
{
	_ASSERTE(pszUserName != NULL);
	_ASSERTE(pszPassword != NULL);

	ULONG lRes;
	CSspiClient Client;
	CSspiServer Server;

	// initialize client object
	lRes = Client.Initialize(pszDomainName, pszUserName, pszPassword);
	if (lRes != ERROR_SUCCESS)
		return SetLastError(lRes), FALSE;

	// initialize server object
	lRes = Server.Initialize();
	if (lRes != ERROR_SUCCESS)
		return SetLastError(lRes), FALSE;

	PVOID pRequest = NULL, pResponse = NULL;
	ULONG cbRequest, cbResponse;

	// start authentication
	lRes = Client.Start(&pRequest, &cbRequest);
	if (lRes != ERROR_MORE_DATA)
		return SetLastError(lRes), FALSE;

	// main request-response loop
	for (;;)
	{
		// process client request and generate a response
		lRes = Server.Continue(pRequest, cbRequest, 
						       &pResponse, &cbResponse);
		if (lRes != ERROR_MORE_DATA)
			break;

		Client.FreeBuffer(pRequest);
		pRequest = NULL;

		// analyze server response and possibly provide more request
		// information
		lRes = Client.Continue(pResponse, cbResponse,
							   &pRequest, &cbRequest);
		if (lRes != ERROR_SUCCESS &&
			lRes != ERROR_MORE_DATA)
			break;

		Server.FreeBuffer(pResponse);
		pResponse = NULL;
	}

	if (pRequest != NULL)
		Client.FreeBuffer(pRequest);
	if (pResponse != NULL)
		Server.FreeBuffer(pResponse);

	if (lRes != ERROR_SUCCESS)
		return SetLastError(lRes), FALSE;

	// check for guest account
	lRes = Server.CheckGuest();
	if (lRes != ERROR_SUCCESS)
		return SetLastError(lRes), FALSE;

	if (phToken != NULL)
	{
		lRes = Server.GetToken(phToken, TOKEN_ALL_ACCESS);
		if (lRes != ERROR_SUCCESS)
			return SetLastError(lRes), FALSE;
	}

	return TRUE;
}

//---------------------------------------------------------------------------
// CheckPassword_ChangePwd
//
//  "Validates" user credentials using NetUserChangePassword function.
//
//  Parameters:
//	  pszDomainName - domain name
//	  pszUserName   - user name
//	  pszPassword	- password
//	  phToken       - pointer to a variable that receives user token
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
CheckPassword_ChangePwd(
	IN PCTSTR pszDomainName,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPassword,
	OUT PHANDLE phToken
	)
{
	_UNUSED(phToken);

	_ASSERTE(pszUserName != NULL);
	_ASSERTE(pszPassword != NULL);
	_ASSERTE(phToken == NULL);

	USES_CONVERSION;

	PCWSTR pszPasswordW = T2CW(pszPassword);

	ULONG lRes = NetUserChangePassword(T2CW(pszDomainName), 
						T2CW(pszUserName), pszPasswordW, pszPasswordW);

	if (lRes == ERROR_INVALID_PASSWORD ||
		lRes == NERR_UserNotFound)
		return SetLastError(ERROR_LOGON_FAILURE), FALSE;

	return TRUE;
}
