// domains.cpp
//
// Trusted domains enumeration.
//
// $Id: $
//

#include "stdafx.h"
#include "domains.h"

#include <lm.h>
#include <ntsecapi.h>
#include <dsgetdc.h>

//---------------------------------------------------------------------------
// EnumTrustedDomains
//
//  Enumerates domains that are trusted to authenticate users by this
//	workstation. The domains are returned by this function without any
//  particular order.
//
//  Parameters:
//	  pEnumProc  - pointer to the callback function
//	  lParam	 - callback function parameter
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
EnumTrustedDomains(
	IN PFNDOMAINNAMEPROC pEnumProc,
	IN LPARAM lParam
	)
{
	_ASSERTE(pEnumProc != NULL);

	HINSTANCE hNetApi32 = GetModuleHandle(_T("netapi32.dll"));
    _ASSERTE(hNetApi32 != NULL);

	DWORD (APIENTRY * pDsEnumerateDomainTrusts)(
		IN PWSTR ServerName,
		IN ULONG Flags,
		OUT PDS_DOMAIN_TRUSTS * Domains,
		OUT	PULONG DomainCount
		);

	NTSTATUS (APIENTRY * pNetEnumerateTrustedDomains)(
		IN PWSTR ServerName,
		OUT PWSTR * DomainNames
		);

#ifdef _UNICODE
    *(FARPROC *)&pDsEnumerateDomainTrusts =
		GetProcAddress(hNetApi32, "DsEnumerateDomainTrustsW");
#else
    *(FARPROC *)&pDsEnumerateDomainTrusts =
		GetProcAddress(hNetApi32, "DsEnumerateDomainTrustsA");
#endif
    *(FARPROC *)&pNetEnumerateTrustedDomains =
		GetProcAddress(hNetApi32, "NetEnumerateTrustedDomains");

	TCHAR szDomainName[DNLEN + 1];
    ULONG lRes;
	BOOL bBreak = FALSE;
    BOOL bAnyDC = FALSE;
    PSERVER_INFO_101 pServerInfo = NULL;

    // on Win2K use DsEnumerateDomainTrustsW function; on earlier systems
    // stick with NetEnumerateTrustedDomains
    if (pDsEnumerateDomainTrusts != NULL)
    {
		PDS_DOMAIN_TRUSTS pTrusts = NULL;
		ULONG cTotal = 0, i;

		// get trusted domains
		lRes = pDsEnumerateDomainTrusts(NULL,
					DS_DOMAIN_IN_FOREST|DS_DOMAIN_DIRECT_OUTBOUND,
					&pTrusts, &cTotal);
		if (lRes == ERROR_SUCCESS)
		{
			// call the user-supplied callback function for each domain
			for (i = 0; !bBreak && i < cTotal; i++)
				bBreak = !pEnumProc(pTrusts[i].NetbiosDomainName, lParam);

			NetApiBufferFree(pTrusts);
		}
	}
    else
    {
		NTSTATUS Status;
		PWSTR pNetDomains = NULL;
		PWSTR psz;

		// retrieve the list of trusted domains
		Status = pNetEnumerateTrustedDomains(NULL, &pNetDomains);
		if (LSA_SUCCESS(Status))
		{
			// call the user-supplied function for each name in the list
			psz = pNetDomains;
			while (!bBreak && *psz != 0)
			{
	#ifdef _UNICODE
				bBreak = !pEnumProc(psz, lParam);
	#else
				WideCharToMultiByte(CP_ACP, 0, psz, -1, szDomainName,
									countof(szDomainName), NULL, NULL);
				bBreak = !pEnumProc(szDomainName, lParam);
	#endif

				psz += wcslen(psz) + 1;
			}

			NetApiBufferFree(pNetDomains);
		}
	}

	if (bBreak)
		return TRUE;

	// none of the above functions enumerates the primary domain, so
	// we have to do it manually
	NTSTATUS Status;
	LSA_HANDLE hPolicy;
	LSA_OBJECT_ATTRIBUTES ObjAttr;
	PPOLICY_PRIMARY_DOMAIN_INFO pDomainInfo;

	memset(&ObjAttr, 0, sizeof(ObjAttr));

	// open security policy object
	Status = LsaOpenPolicy(NULL, &ObjAttr, 
						   POLICY_VIEW_LOCAL_INFORMATION, &hPolicy);
	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;

	// query primary domain information Q179891
	Status = LsaQueryInformationPolicy(hPolicy, 
									   PolicyPrimaryDomainInformation,
									   (PVOID *)&pDomainInfo);

	if (LSA_SUCCESS(Status))
	{
		if (pDomainInfo->Sid != NULL)
		{
#ifdef _UNICODE
			bBreak = !pEnumProc(pDomainInfo->Name.Buffer, lParam);
#else
			WideCharToMultiByte(CP_ACP, 0, pDomainInfo->Name.Buffer, -1,
								szDomainName, countof(szDomainName), NULL, 
								NULL);
			bBreak = !pEnumProc(szDomainName, lParam);
#endif
		}

		LsaFreeMemory(pDomainInfo);
	}

	LsaClose(hPolicy);

	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;
	else if (bBreak)
		return TRUE;

    // determine whether this computer is a primary or backup domain
    // controller
    lRes = NetServerGetInfo(NULL, 101, (PBYTE *)&pServerInfo);
    if (lRes != ERROR_SUCCESS)
		return SetLastError(lRes), FALSE;

	bAnyDC = pServerInfo->sv101_type & 
				(SV_TYPE_DOMAIN_CTRL|SV_TYPE_DOMAIN_BAKCTRL);

	NetApiBufferFree(pServerInfo);

	// if not a DC, then include the name of the local computer into the
	// enumeration
	if (!bAnyDC)
	{
		ULONG cchDomainName = countof(szDomainName);
		_VERIFY(GetComputerName(szDomainName, &cchDomainName));

		pEnumProc(szDomainName, lParam);
	}

	return TRUE;
}
