// privapi.cpp
//
// Privilege manipulation.
//
// $Id: $
//

#include "stdafx.h"
#include "privapi.h"

#include <lm.h>
#include <ntsecapi.h>

// from NTDDK
#define STATUS_SUCCESS					 ((NTSTATUS)0x00000000L)
#define STATUS_OBJECT_NAME_NOT_FOUND     ((NTSTATUS)0xC0000034L)

//---------------------------------------------------------------------------
// IsUserAdmin
//
//  Checks whether the current user is a member of the local administrators
//  group.
//
//  Parameters:
//	  none.
//
//  Returns:
//	  TRUE, if the user is a member of the local administrators group,
//	  FALSE - if it is not, or if an error has occured.
//
EXTERN_C
BOOL
APIENTRY
IsUserAdmin()
{
	HANDLE hToken;
	DWORD cbNeeded;
	PTOKEN_GROUPS pGroups;

	// obtain the token of the current thread 
	if (!OpenThreadToken(GetCurrentThread(), TOKEN_QUERY, FALSE, &hToken))
	{
		if (GetLastError() != ERROR_NO_TOKEN)
			return FALSE;

		// revert to the process token
		if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
			return FALSE;
	}

	// determine size of the buffer required to receive all groups
	if (!GetTokenInformation(hToken, TokenGroups, NULL, 0, &cbNeeded))
	{
		if (GetLastError() != ERROR_INSUFFICIENT_BUFFER)
		{
			DWORD dwError = GetLastError();
			_VERIFY(CloseHandle(hToken));
			return SetLastError(dwError), FALSE;
		}
	}

	// allocate privileges buffer
	pGroups = (PTOKEN_GROUPS)_alloca(cbNeeded);
	_ASSERTE(pGroups != NULL);

	// retrieve token privileges
	if (!GetTokenInformation(hToken, TokenGroups, pGroups, cbNeeded,
							 &cbNeeded))
	{
		DWORD dwError = GetLastError();
		_VERIFY(CloseHandle(hToken));
		return SetLastError(dwError), FALSE;
	}

	_VERIFY(CloseHandle(hToken));

    SID_IDENTIFIER_AUTHORITY siaNT = SECURITY_NT_AUTHORITY;
    PSID pSidAdmins = NULL;
	BOOL bAdmin = FALSE;

	// create well-known SID of the local administrators group
	if (!AllocateAndInitializeSid(&siaNT, 2, SECURITY_BUILTIN_DOMAIN_RID,
					DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, &pSidAdmins))
		return FALSE;

	// walk through all groups and check if the local administrators group
	// is present in the caller's token
	for (ULONG i = 0; !bAdmin && i < pGroups->GroupCount; i++)
		bAdmin = EqualSid(pSidAdmins, pGroups->Groups[i].Sid);
	
	FreeSid(pSidAdmins);

	SetLastError(0);
	return bAdmin;
}

//---------------------------------------------------------------------------
// IsPrivilege
//
//  Checks if the privilege is present in the caller's token.
//
//  Parameters:
//	  pszPrivilegeName - name of the privilege
//
//  Returns:
//	  TRUE, if the privilege is present, FALSE - if it isn't, or if an
//	  error has occured.
//
EXTERN_C
BOOL
APIENTRY
IsPrivilege(
	IN PCTSTR pszPrivilegeName
	)
{
	_ASSERTE(pszPrivilegeName != NULL);

	LUID Luid;
	_VERIFY(LookupPrivilegeValue(NULL, pszPrivilegeName, &Luid));

	HANDLE hToken;
	DWORD cbNeeded;
	PTOKEN_PRIVILEGES pPriv;

	// obtain the token of the current thread 
	if (!OpenThreadToken(GetCurrentThread(), TOKEN_QUERY, FALSE, &hToken))
	{
		if (GetLastError() != ERROR_NO_TOKEN)
			return FALSE;

		// revert to the process token
		if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
			return FALSE;
	}

	// determine size of the buffer required to receive all privileges
	if (!GetTokenInformation(hToken, TokenPrivileges, NULL, 0, &cbNeeded))
	{
		if (GetLastError() != ERROR_INSUFFICIENT_BUFFER)
		{
			DWORD dwError = GetLastError();
			_VERIFY(CloseHandle(hToken));
			return SetLastError(dwError), FALSE;
		}
	}

	// allocate privileges buffer
	pPriv = (PTOKEN_PRIVILEGES)_alloca(cbNeeded);
	_ASSERTE(pPriv != NULL);

	// retrieve token privileges
	if (!GetTokenInformation(hToken, TokenPrivileges, pPriv, cbNeeded,
							 &cbNeeded))
	{
		DWORD dwError = GetLastError();
		_VERIFY(CloseHandle(hToken));
		return SetLastError(dwError), FALSE;
	}

	_VERIFY(CloseHandle(hToken));

	// walk through all privileges and check if the specified privilege
	// is present in the caller's token
	for (UINT i = 0; i < pPriv->PrivilegeCount; i++)
	{
		if (pPriv->Privileges[i].Luid.LowPart == Luid.LowPart &&
			pPriv->Privileges[i].Luid.HighPart == Luid.HighPart)
			return TRUE;
	}

	SetLastError(ERROR_SUCCESS);
	return FALSE;
}

//---------------------------------------------------------------------------
// InitUnicodeString
//
//  Initializes an LSA_UNICODE_STRING structure.
//
//  Parameters:
//	  pUnicodeString - pointer to the structure
//	  pSourceString  - pointer to a null-terminated string
//
//  Returns:
//	  no return value.
//
VOID
APIENTRY
InitUnicodeString(
	OUT PLSA_UNICODE_STRING pUnicodeString,
	IN PCWSTR pSourceString
	)
{
	_ASSERTE(pUnicodeString != NULL);
	_ASSERTE(pSourceString != NULL);

	USHORT Length = (USHORT)(wcslen(pSourceString) * sizeof(WCHAR));

	pUnicodeString->Length		  = Length;
	pUnicodeString->MaximumLength = (USHORT)(Length + sizeof(WCHAR));
	pUnicodeString->Buffer		  = (PWSTR)pSourceString;
}

//---------------------------------------------------------------------------
// EqualUnicodeString
//
//  Compares two counted strings.
//
//  Parameters:
//	  pString1 - pointer to the first string
//	  pString2 - pointer to the second string
//
//  Returns:
//	  TRUE, if the strings are equal, FALSE - otherwise.
//
BOOL
APIENTRY
EqualUnicodeString(
	IN CONST LSA_UNICODE_STRING * pString1,
	IN CONST LSA_UNICODE_STRING * pString2
	)
{
	_ASSERTE(pString1 != NULL);
	_ASSERTE(pString2 != NULL);

	return pString1->Length == pString2->Length
		   && !memcmp(pString1->Buffer, pString2->Buffer, pString1->Length);
}

//---------------------------------------------------------------------------
// GrantPrivilege
//
//  Grants the specified privilege to the specified user account.
//
//  Parameters:
//	  pszMachineName   - name of the target machine
//	  pUserSid		   - user SID
//	  pszUserName	   - user name
//	  pszPrivilegeName - name of the privilege
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
GrantPrivilege(
	IN PCTSTR pszMachineName,
	IN PSID pUserSid,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPrivilegeName
	)
{
	_ASSERTE(pUserSid != NULL || pszUserName != NULL);
	_ASSERTE(pUserSid == NULL || IsValidSid(pUserSid));
	_ASSERTE(pszPrivilegeName != NULL);

	BYTE bSid[8 + 4 * SID_MAX_SUB_AUTHORITIES];

	if (pUserSid == NULL)
	{
		DWORD cbSid = sizeof(bSid);
		TCHAR szDomainName[DNLEN + 1];
		DWORD cchDomainName = countof(szDomainName);
		SID_NAME_USE Use;

		if (pszUserName[0] == _T('.') &&
			pszUserName[1] == _T('\\'))
			pszUserName += 2;

		// lookup user account SID
		if (!LookupAccountName(pszMachineName, pszUserName, (PSID)bSid,
							   &cbSid, szDomainName, &cchDomainName, &Use))
			return FALSE;

		pUserSid = (PSID)bSid;
	}

	_ASSERTE(pUserSid != NULL);
	_ASSERTE(IsValidSid(pUserSid));

	USES_CONVERSION;

	NTSTATUS Status;
	LSA_HANDLE hPolicy;
	LSA_UNICODE_STRING SystemName;
	LSA_UNICODE_STRING UserRight;
	PLSA_UNICODE_STRING pSystemName = NULL;
	LSA_OBJECT_ATTRIBUTES ObjAttr;

	if (pszMachineName != NULL)
	{
		InitUnicodeString(&SystemName, T2CW(pszMachineName));
		pSystemName = &SystemName;
	}

	InitUnicodeString(&UserRight, T2CW(pszPrivilegeName));
	memset(&ObjAttr, 0, sizeof(ObjAttr));

	// open policy object
	Status = LsaOpenPolicy(pSystemName, &ObjAttr, 
						   POLICY_LOOKUP_NAMES|POLICY_CREATE_ACCOUNT,
						   &hPolicy);
	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;

	// add privilege to the account
	Status = LsaAddAccountRights(hPolicy, pUserSid, &UserRight, 1);

	// close policy handle
	LsaClose(hPolicy);

	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;

	return TRUE;
}

//---------------------------------------------------------------------------
// RevokePrivilege
//
//  Revokes the specified privilege from the specified user account.
//
//  Parameters:
//	  pszMachineName   - name of the target machine
//	  pUserSid		   - user SID
//	  pszUserName	   - user name
//	  pszPrivilegeName - name of the privilege
//
//  Returns:
//	  TRUE, if successful, FALSE - otherwise.
//
EXTERN_C
BOOL
APIENTRY
RevokePrivilege(
	IN PCTSTR pszMachineName,
	IN PSID pUserSid,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPrivilegeName
	)
{
	_ASSERTE(pUserSid != NULL || pszUserName != NULL);
	_ASSERTE(pUserSid == NULL || IsValidSid(pUserSid));
	_ASSERTE(pszPrivilegeName != NULL);

	BYTE bSid[8 + 4 * SID_MAX_SUB_AUTHORITIES];

	if (pUserSid == NULL)
	{
		DWORD cbSid = sizeof(bSid);
		TCHAR szDomainName[DNLEN + 1];
		DWORD cchDomainName = countof(szDomainName);
		SID_NAME_USE Use;

		if (pszUserName[0] == _T('.') &&
			pszUserName[1] == _T('\\'))
			pszUserName += 2;

		// lookup user account SID
		if (!LookupAccountName(pszMachineName, pszUserName, (PSID)bSid,
							   &cbSid, szDomainName, &cchDomainName, &Use))
			return FALSE;

		pUserSid = (PSID)bSid;
	}

	_ASSERTE(pUserSid != NULL);
	_ASSERTE(IsValidSid(pUserSid));

	USES_CONVERSION;

	NTSTATUS Status;
	LSA_HANDLE hPolicy;
	LSA_UNICODE_STRING SystemName;
	LSA_UNICODE_STRING UserRight;
	PLSA_UNICODE_STRING pSystemName = NULL;
	LSA_OBJECT_ATTRIBUTES ObjAttr;

	if (pszMachineName != NULL)
	{
		InitUnicodeString(&SystemName, T2CW(pszMachineName));
		pSystemName = &SystemName;
	}

	InitUnicodeString(&UserRight, T2CW(pszPrivilegeName));
	memset(&ObjAttr, 0, sizeof(ObjAttr));

	// open policy object
	Status = LsaOpenPolicy(pSystemName, &ObjAttr, POLICY_LOOKUP_NAMES,
						   &hPolicy);
	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;

	// remove privilege from the account
	Status = LsaRemoveAccountRights(hPolicy, pUserSid, FALSE, &UserRight, 1);
	if (Status == STATUS_OBJECT_NAME_NOT_FOUND)
		Status = STATUS_SUCCESS;

	// close policy handle
	LsaClose(hPolicy);

	if (!LSA_SUCCESS(Status))
		return SetLastError(LsaNtStatusToWinError(Status)), FALSE;

	return TRUE;
}
