// sspiauth.cpp
//
// SSPI authentication.
//
// $Id: $
//

#include "stdafx.h"
#include "sspiauth.h"

#include <lm.h>
#include <ntsecapi.h>

//---------------------------------------------------------------------------
// CSspiBase
//
//  Constructor, initializes the object.
//
//  Parameters:
//	  none.
//
//  Returns:
//    no return value.
//
CSspiBase::CSspiBase()
{
	m_hSecDll     = NULL;
	m_pEntryTable = NULL;
	m_cbMaxToken  = 0;

    SecInvalidateHandle(&m_hCtx);
    SecInvalidateHandle(&m_hCred);	

	m_bHaveCredHandle  = false;
	m_bHaveCtxHandle   = false;
}

//---------------------------------------------------------------------------
// ~CSspiBase
//
//  Destructor, performs necessary cleanup.
//
//  Parameters:
//	  none.
//
//  Returns:
//	  no return value.
//
CSspiBase::~CSspiBase()
{
    if (m_bHaveCtxHandle)
		m_pEntryTable->DeleteSecurityContext(&m_hCtx);

    if (m_bHaveCredHandle)
		m_pEntryTable->FreeCredentialHandle(&m_hCred);

    if (m_hSecDll != NULL)
		_VERIFY(FreeLibrary(m_hSecDll));
}

//---------------------------------------------------------------------------
// Initialize
//
//  Initializes the security package.
//
//  Parameters:
//	  pszProvider - package name
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiBase::Initialize(
	IN PCTSTR pszProvider
	)
{
	_UNUSED(pszProvider);

	_ASSERTE(pszProvider != NULL);

    // load security library
    m_hSecDll = LoadLibrary(_T("secur32.dll"));
    if (m_hSecDll == NULL)
		m_hSecDll = LoadLibrary(_T("security.dll"));
    if (m_hSecDll == NULL)
    {
		_RPT_API_FAILED(LoadLibrary);
		return GetLastError();
    }

    // find address of the initialization function
    FARPROC pfnInit = GetProcAddress(m_hSecDll, SECURITY_ENTRYPOINT_ANSI);
    if (pfnInit == NULL)
    {
		_RPT_API_FAILED(GetProcAddress);
		return GetLastError();
    }

    // obtain address of the function table
    m_pEntryTable = (PSecurityFunctionTable)(*pfnInit)();
    if (m_pEntryTable == NULL)
		return ERROR_GEN_FAILURE;

    PSecPkgInfo pInfo = NULL;

    // query for the package we are interested in
    SECURITY_STATUS Status = 
		m_pEntryTable->QuerySecurityPackageInfo((PWSTR)pszProvider, &pInfo);
    if (!NT_SUCCESS(Status))
		return LsaNtStatusToWinError(Status);

    // determine max token length
    m_cbMaxToken = pInfo->cbMaxToken;
    m_pEntryTable->FreeContextBuffer(pInfo);

    return ERROR_SUCCESS;
}

//---------------------------------------------------------------------------
// CompleteToken
//
//  This is a helper function that completes authentication token and 
//  returns token buffer.
//
//  Parameters:
//	  Status   - status value
//	  pBufDesc - token buffer
//	  ppToken  - pointer to a variable that receives token buffer pointer
//	  pcbToken - pointer to a variable that recevies token size in bytes
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiBase::CompleteToken(
	IN SECURITY_STATUS Status,
	IN PSecBufferDesc pBufDesc,
	OUT PVOID * ppToken,
	OUT PULONG pcbToken
	)
{
	BOOL bContinue = FALSE;
	PVOID pToken = pBufDesc->pBuffers->pvBuffer;

	switch (Status)
	{
		case SEC_I_CONTINUE_NEEDED:
			bContinue = TRUE;
			break;

		case SEC_I_COMPLETE_AND_CONTINUE:
			bContinue = TRUE;
			// fall through

		case SEC_I_COMPLETE_NEEDED:
			if (m_pEntryTable->CompleteAuthToken == NULL)
			{
				_RPT0(_CRT_WARN, "Cannot complete authentication token,"
								 " CompleteAuthToken entry not found\n");
				FreeBuffer(pToken);
				return ERROR_CALL_NOT_IMPLEMENTED;
			}

			Status = m_pEntryTable->CompleteAuthToken(&m_hCtx, pBufDesc);
			break;

		default:
			break;
	}

	if (!NT_SUCCESS(Status))
	{
		FreeBuffer(pToken);
		return LsaNtStatusToWinError(Status);
	}

    *ppToken = pToken;
    *pcbToken = pBufDesc->pBuffers->cbBuffer;

    return bContinue ? ERROR_MORE_DATA : ERROR_SUCCESS;
}

//---------------------------------------------------------------------------
// Initialize
//
//  Initializes the object, preparing it to handle authentication of the
//  specified user.
//
//  Parameters:
//	  pszDomainName - domain name
//	  pszUserName   - user name
//	  pszPassword   - password
//	  pszProvider   - security package name
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiClient::Initialize(
	IN PCTSTR pszDomainName,
	IN PCTSTR pszUserName,
	IN PCTSTR pszPassword,
	IN PCTSTR pszProvider
	)
{
	_ASSERTE(pszUserName != NULL);
	_ASSERTE(pszPassword != NULL);
	_ASSERTE(pszProvider != NULL);

	// initialize security package
	ULONG lRes = CSspiBase::Initialize(pszProvider);
	if (lRes != ERROR_SUCCESS)
		return lRes;

	SECURITY_STATUS Status;
    TimeStamp Lifetime;
    SEC_WINNT_AUTH_IDENTITY Id;

	Id.Domain         = (PWSTR)pszDomainName;
	Id.DomainLength   = pszDomainName == NULL ? 0 : lstrlen(pszDomainName);
	Id.User           = (PWSTR)pszUserName;
	Id.UserLength     = lstrlen(pszUserName);
	Id.Password		  = (PWSTR)pszPassword;
	Id.PasswordLength = lstrlen(pszPassword);

#ifdef _UNICODE
    Id.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE;
#else
    Id.Flags = SEC_WINNT_AUTH_IDENTITY_ANSI;
#endif

	// initialize credentials handle
    Status = m_pEntryTable->AcquireCredentialsHandle(
							NULL, (PWSTR)pszProvider, SECPKG_CRED_OUTBOUND,
							NULL, &Id, NULL, NULL, &m_hCred, &Lifetime);

    if (!NT_SUCCESS(Status))
	{
		_RPT_API_FAILED_EX(AcquireCredentialsHandle, Status);
		return LsaNtStatusToWinError(Status);
	}

    m_bHaveCredHandle = true;
    return ERROR_SUCCESS;
}

//---------------------------------------------------------------------------
// Start
//
//  Generates initial authentication request.
//
//  Parameters:
//	  ppRequest  - pointer to a variable that receives request buffer pointer
//	  pcbRequest - pointer to a variable that receives request size in bytes
//
//  Returns:
//	  Win32 error code.
//
ULONG 
CSspiClient::Start(
	OUT PVOID * ppRequest,
	OUT PULONG pcbRequest
	)
{
	_ASSERTE(ppRequest != NULL);
	_ASSERTE(pcbRequest != NULL);
	_ASSERTE(m_bHaveCredHandle);

	*ppRequest = NULL;
	*pcbRequest = NULL;

    SECURITY_STATUS Status;
    SecBuffer OutBuf;
    SecBufferDesc OutBufDesc;

    // prepare output buffer
    OutBuf.cbBuffer = m_cbMaxToken;
    OutBuf.BufferType = SECBUFFER_TOKEN;
    OutBuf.pvBuffer = (PVOID)LocalAlloc(LMEM_FIXED, m_cbMaxToken);

	if (OutBuf.pvBuffer == NULL)
		return ERROR_NOT_ENOUGH_MEMORY;

    OutBufDesc.ulVersion = 0;
    OutBufDesc.cBuffers = 1;
    OutBufDesc.pBuffers = &OutBuf;

    TimeStamp Lifetime;
    ULONG ContextAttributes;

    // initialize security context
    Status = m_pEntryTable->InitializeSecurityContext(
						&m_hCred, NULL, NULL, 0, 0,  SECURITY_NATIVE_DREP, 
						NULL, 0, &m_hCtx, &OutBufDesc, &ContextAttributes, 
						&Lifetime);
    if (!NT_SUCCESS(Status))
	{
		_RPT_API_FAILED_EX(InitializeSecurityContext, Status);

		LocalFree((HLOCAL)OutBuf.pvBuffer);
		return LsaNtStatusToWinError(Status);
	}

    m_bHaveCtxHandle = true;
	return CompleteToken(Status, &OutBufDesc, ppRequest, pcbRequest);
}

//---------------------------------------------------------------------------
// Continue
//
//  Continues authentication process. Processes server response and provides
//	more request data if necessary.
//
//  Parameters:
//	  pResponse  - pointer to the respose buffer
//	  cbResponse - size of the response buffer
//	  ppRequest  - pointer to a variable that receives request buffer pointer
//	  pcbRequest - pointer to a variable that receives request size in bytes
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiClient::Continue(
	IN PVOID pResponse,
	IN ULONG cbResponse,
	OUT PVOID * ppRequest,
	OUT PULONG pcbRequest
	)
{
	_ASSERTE(pResponse != NULL);
	_ASSERTE(cbResponse > 0);
	_ASSERTE(ppRequest != NULL);
	_ASSERTE(pcbRequest != NULL);

	_ASSERTE(m_bHaveCredHandle);
	_ASSERTE(m_bHaveCtxHandle);

    SECURITY_STATUS Status;
    SecBuffer InBuf, OutBuf;
    SecBufferDesc InBufDesc, OutBufDesc;

    // prepare input buffer
    InBuf.cbBuffer = cbResponse;
    InBuf.BufferType = SECBUFFER_TOKEN;
    InBuf.pvBuffer = pResponse;

    InBufDesc.ulVersion = 0;
    InBufDesc.cBuffers = 1;
    InBufDesc.pBuffers = &InBuf;

    // prepare output buffer
    OutBuf.cbBuffer = m_cbMaxToken;
    OutBuf.BufferType = SECBUFFER_TOKEN;
    OutBuf.pvBuffer = (PVOID)LocalAlloc(LMEM_FIXED, m_cbMaxToken);

	if (OutBuf.pvBuffer == NULL)
		return ERROR_NOT_ENOUGH_MEMORY;

    OutBufDesc.ulVersion = 0;
    OutBufDesc.cBuffers = 1;
    OutBufDesc.pBuffers = &OutBuf;

    TimeStamp Lifetime;
    ULONG ContextAttributes;

    // initialize security context
    Status = m_pEntryTable->InitializeSecurityContext(
						&m_hCred, &m_hCtx, NULL, 0, 0, SECURITY_NATIVE_DREP, 
						&InBufDesc, 0, &m_hCtx, &OutBufDesc, 
						&ContextAttributes, &Lifetime);
    if (!NT_SUCCESS(Status))
	{
		_RPT_API_FAILED_EX(InitializeSecurityContext, Status);

		LocalFree((HLOCAL)OutBuf.pvBuffer);
		return LsaNtStatusToWinError(Status);
	}

	return CompleteToken(Status, &OutBufDesc, ppRequest, pcbRequest);
}

//---------------------------------------------------------------------------
// Initialize
//
//  Initializes the object preparing it to handle authentication request.
//
//  Parameters:
//	  pszProvider - security package name
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiServer::Initialize(
	IN PCTSTR pszProvider
	)
{
	_ASSERTE(pszProvider != NULL);

	// initialize security package
	ULONG lRes = CSspiBase::Initialize(pszProvider);
	if (lRes != ERROR_SUCCESS)
		return lRes;

    SECURITY_STATUS Status;
    TimeStamp Lifetime;

    Status = m_pEntryTable->AcquireCredentialsHandle(
							NULL, (PWSTR)pszProvider, SECPKG_CRED_INBOUND,
							NULL, NULL, NULL, NULL, &m_hCred,
							&Lifetime);

    if (!NT_SUCCESS(Status))
	{
		_RPT_API_FAILED_EX(AcquireCredentialsHandle, Status);
		return LsaNtStatusToWinError(Status);
	}

    m_bHaveCredHandle = true;
    return ERROR_SUCCESS;
}

//---------------------------------------------------------------------------
// Continue
//
//  Handles authentication request from a client
//
//  Parameters:
//    pRequest	  - pointer to the request buffer
//	  cbRequest   - size of the request in bytes
//	  ppResponse  - pointer to a variable that receives response buffer
//					pointer
//	  pcbResponse - pointer to a variable that receives size of the
//					response in bytes
//
//  Returns:
//    * ERROR_SUCCESS, if authentication is completed successfully;
//    * ERROR_MORE_DATA, more roundtrips are required to complete authentica-
//	    tion and a new response is returned;
//    * ERROR_LOGON_FAILURE, if authentication failed, no response data
//	    is produced;
//    * other values indicate errors and no response is returned in such
//	    cases.
//
ULONG
CSspiServer::Continue(
	IN PVOID pRequest,
	IN ULONG cbRequest,
	OUT PVOID * ppResponse,
	OUT PULONG pcbResponse
	)
{
	_ASSERTE(pRequest != NULL);
	_ASSERTE(cbRequest != NULL);
	_ASSERTE(ppResponse != NULL);
	_ASSERTE(pcbResponse != NULL);

	*ppResponse = NULL;
	*pcbResponse = 0;

    SECURITY_STATUS Status;
    SecBuffer InBuf, OutBuf;
    SecBufferDesc InBufDesc, OutBufDesc;

    // prepare input buffer
    InBuf.cbBuffer = cbRequest;
    InBuf.BufferType = SECBUFFER_TOKEN;
    InBuf.pvBuffer = pRequest;

    InBufDesc.ulVersion = 0;
    InBufDesc.cBuffers = 1;
    InBufDesc.pBuffers = &InBuf;

    // prepare output buffer
    OutBuf.cbBuffer = m_cbMaxToken;
    OutBuf.BufferType = SECBUFFER_TOKEN;
    OutBuf.pvBuffer = (PVOID)LocalAlloc(LMEM_FIXED, m_cbMaxToken);

	if (OutBuf.pvBuffer == NULL)
		return ERROR_NOT_ENOUGH_MEMORY;

    OutBufDesc.ulVersion = 0;
    OutBufDesc.cBuffers = 1;
    OutBufDesc.pBuffers = &OutBuf;

    TimeStamp Lifetime;
    ULONG ContextAttributes;

    // accept security context
    Status = m_pEntryTable->AcceptSecurityContext(
						&m_hCred, m_bHaveCtxHandle ? &m_hCtx : NULL,
						&InBufDesc, 0,  SECURITY_NATIVE_DREP, &m_hCtx,  
						&OutBufDesc,  &ContextAttributes, &Lifetime);
    if (!NT_SUCCESS(Status))
	{
		_RPT_API_FAILED_EX(AcceptSecurityContext, Status);
		
		LocalFree((HLOCAL)OutBuf.pvBuffer);
		return LsaNtStatusToWinError(Status);
	}

    m_bHaveCtxHandle = true;
	return CompleteToken(Status, &OutBufDesc, ppResponse, pcbResponse);
}

//---------------------------------------------------------------------------
// GetToken
//
//  Returns authentication token.
//
//  Parameters:
//	  phToken  - pointer to a variable that receives the token
//	  dwAccess - token access rights
//
//  Returns:
//	  no return value.
//
ULONG
CSspiServer::GetToken(
	OUT PHANDLE phToken,
	IN ULONG dwAccess
	)
{
	_ASSERTE(phToken != NULL);
	_ASSERTE(m_bHaveCtxHandle);

	*phToken = NULL;

	SECURITY_STATUS Status;

	// impersonate security context
	Status = m_pEntryTable->ImpersonateSecurityContext(&m_hCtx);
	if (!NT_SUCCESS(Status))
		return LsaNtStatusToWinError(Status);

	// obtain token associated with the current thread
	if (!OpenThreadToken(GetCurrentThread(), dwAccess, FALSE, phToken))
	{
		Status = GetLastError();
		m_pEntryTable->RevertSecurityContext(&m_hCtx);
		return (ULONG)Status;
	}

	m_pEntryTable->RevertSecurityContext(&m_hCtx);
	return ERROR_SUCCESS;
}

//---------------------------------------------------------------------------
// CheckGuest
//
//  Checks for guest account.
//
//  Parameters:
//	  none.
//
//  Returns:
//	  Win32 error code.
//
ULONG
CSspiServer::CheckGuest()
{
	_ASSERTE(m_bHaveCtxHandle);

	SECURITY_STATUS Status;
	SecPkgContext_Names Names;

	// query the name of the user
	Status = m_pEntryTable->QueryContextAttributes(&m_hCtx,
								SECPKG_ATTR_NAMES, &Names);
	if (!NT_SUCCESS(Status))
		return LsaNtStatusToWinError(Status);

	PTSTR pszDomainName = Names.sUserName;
	PTSTR pszUserName = _tcschr(pszDomainName, _T('\\'));
	HANDLE hToken = NULL;

	BYTE bUser[sizeof(TOKEN_USER) + 8 + 4 * SID_MAX_SUB_AUTHORITIES];
	PTOKEN_USER pUser = (PTOKEN_USER)bUser;
	ULONG cbNeeded;

	TCHAR szUserName[UNLEN + 1];
	TCHAR szDomainName[DNLEN + 1];
	ULONG cbUserName = countof(szUserName);
	ULONG cbDomainName = countof(szDomainName);
	SID_NAME_USE Use;
	
	for (;;)
	{
		if (pszUserName == NULL)
		{
			Status = ERROR_LOGON_FAILURE;
			break;
		}
		*pszUserName++ = 0;

		// obtain logged on user token
		Status = GetToken(&hToken, TOKEN_QUERY);
		if (Status != ERROR_SUCCESS)
			break;

		// retrieve user sid associated with the token
		if (!GetTokenInformation(hToken, TokenUser, bUser, sizeof(bUser),
								 &cbNeeded))
		{
			_RPT_API_FAILED(GetTokenInformation);

			Status = GetLastError();
			break;
		}

		// retrieve the actual name of the logged on user
		if (!LookupAccountSid(NULL, pUser->User.Sid, szUserName, 
							  &cbUserName, szDomainName, &cbDomainName,
							  &Use))
		{
			_RPT_API_FAILED(LookupAccountSid);

			Status = GetLastError();
			break;
		}

		if (lstrcmpi(pszDomainName, szDomainName) != 0 ||
			lstrcmpi(pszUserName, szUserName) != 0)
		{
			Status = ERROR_LOGON_FAILURE;
			break;
		}

		Status = ERROR_SUCCESS;
		break;
	}

	if (hToken != NULL)
		_VERIFY(CloseHandle(hToken));

	m_pEntryTable->FreeContextBuffer(Names.sUserName);
	return Status;
}
